@@ -188,6 +188,7 @@ def train_on_responses_only(
     force_match      = True,  # Match newlines as well!
     tokenizer        = None,  # Optional
     return_function  = False, # Useful for iterating over lists
+    num_proc         = None,
 ):
     """
     Trains only on responses and not on the instruction by masking out
@@ -328,25 +329,34 @@ def train_on_responses_only(
         return _train_on_responses_only
 
     from multiprocessing import cpu_count
-    num_proc = cpu_count()
+    if num_proc is None or type(num_proc) is not int: num_proc = cpu_count()
 
     if hasattr(trainer, "train_dataset") and trainer.train_dataset is not None:
         if not hasattr(trainer.train_dataset, "map"):
             raise TypeError("Unsloth: train_on_responses_only does not work on lists!")
-        trainer.train_dataset = trainer.train_dataset.map(_train_on_responses_only, batched = True, num_proc = num_proc)
+        if isinstance(trainer.train_dataset, IterableDataset):
+            trainer.train_dataset = trainer.train_dataset.map(_train_on_responses_only, batch_size = trainer.train_dataset._ex_iterable.batch_size, batched = True)
+        else:
+            trainer.train_dataset = trainer.train_dataset.map(_train_on_responses_only, batched = True, num_proc = num_proc)
     pass
     
-    if hasattr(trainer, "eval_dataset")  and trainer.eval_dataset  is not None:
+    if hasattr(trainer, "eval_dataset") and trainer.eval_dataset is not None:
         # Eval datasets could be a dict!
         if type(trainer.eval_dataset) is dict:
             for key, value in trainer.eval_dataset.items():
                 if not hasattr(value, "map"):
                     raise TypeError("Unsloth: train_on_responses_only does not work on lists!")
-                trainer.eval_dataset[key] = value.map(_train_on_responses_only, batched = True, num_proc = num_proc)
+                if isinstance(value, IterableDataset):
+                    trainer.eval_dataset[key] = value.map(_train_on_responses_only, batch_size = value._ex_iterable.batch_size, batched = True)
+                else:
+                    trainer.eval_dataset[key] = value.map(_train_on_responses_only, batched = True, num_proc = num_proc)
         else:
             if not hasattr(trainer.eval_dataset, "map"):
                 raise TypeError("Unsloth: train_on_responses_only does not work on lists!")
-            trainer.eval_dataset = trainer.eval_dataset.map(_train_on_responses_only, batched = True, num_proc = num_proc)
+            if isinstance(trainer.eval_dataset, IterableDataset):
+                trainer.eval_dataset = trainer.eval_dataset.map(_train_on_responses_only, batch_size = trainer.eval_dataset._ex_iterable.batch_size, batched = True)
+            else:
+                trainer.eval_dataset = trainer.eval_dataset.map(_train_on_responses_only, batched = True, num_proc = num_proc)
         pass
     pass
 
@@ -467,7 +477,12 @@ pass
 
 
 from datasets import (Dataset, IterableDataset,)
-from trl.trainer.utils import ConstantLengthDataset
+try:
+    from trl.trainer.utils import ConstantLengthDataset
+except:
+    # TRL 0.20.0 removes ConstantLengthDataset
+    ConstantLengthDataset = None
+
 # Faster SFTTrainer prepare_dataset
 def sft_prepare_dataset(
     self,
@@ -479,7 +494,10 @@ def sft_prepare_dataset(
     dataset_name: str,
 ) -> Union[Dataset, IterableDataset]:
     # All Unsloth Zoo code licensed under LGPLv3
-    if isinstance(dataset, ConstantLengthDataset): return dataset
+    try:
+        if isinstance(dataset, ConstantLengthDataset): return dataset
+    except:
+        pass
 
     map_kwargs = {}
     use_desc = isinstance(dataset, Dataset)
@@ -492,6 +510,7 @@ def sft_prepare_dataset(
     if max_seq_length == 0: max_seq_length = getattr(args, "max_seq_length", 0)
     if max_seq_length == 0: max_seq_length = getattr(self, "max_seq_length", 0)
     if max_seq_length == 0: max_seq_length = getattr(self, "max_seq", 0)
+    if max_seq_length == 0: raise RuntimeError("Unsloth: max_seq_length is 0! Please specify one!")
     dataset_text_field = getattr(args, "dataset_text_field", "text")
     do_truncation = max_seq_length != 0
     do_formatting_func = False
@@ -529,14 +548,14 @@ def sft_prepare_dataset(
     if do_tokenize:
         # Check double BOS tokens
         if do_formatting_func:
-            test_text = formatting_func(dataset[0])
+            test_text = formatting_func(next(iter(dataset)))
             if not isinstance(test_text, list):
                 raise ValueError(
                     "Unsloth: The `formatting_func` should return a list of processed strings."
                 )
             test_text = test_text[0]
         else:
-            test_text = dataset[0][dataset_text_field]
+            test_text = next(iter(dataset))[dataset_text_field][0]
 
         # Get chat template
         chat_template = getattr(processing_class, 'chat_template', '')
@@ -568,7 +587,11 @@ def sft_prepare_dataset(
             )
         pass
 
-        map_kwargs["num_proc"] = getattr(args, "dataset_num_proc", 2)
+        if not isinstance(dataset, IterableDataset):
+            map_kwargs["num_proc"] = getattr(args, "dataset_num_proc", 2)
+        else:
+            map_kwargs["batch_size"] = dataset._ex_iterable.batch_size
+            
         if use_desc: map_kwargs["desc"] = f'Unsloth: Tokenizing ["{dataset_text_field}"]'
         dataset = dataset.map(_tokenize, batched = True, **map_kwargs)
 
