@@ -16,6 +16,8 @@
 
 import torch
 import os
+import re
+import ast
 
 __all__ = [
     "patch_compiling_bitsandbytes",
@@ -26,23 +28,31 @@ __all__ = [
 ]
 
 from .compiler import UNSLOTH_COMPILE_LOCATION
-from .utils import _get_dtype
+from .utils import _get_dtype, Version
 
 # Also disable compiling on bitsandbytes
 def patch_compiling_bitsandbytes():
     # All Unsloth Zoo code licensed under LGPLv3
     os.environ["UNSLOTH_PATCHED"] = "1"
 
-    # Disable dynamo on Linear4bit, Linear8bit and other future modules
-    for x in ["bitsandbytes.nn.modules", "peft.tuners.lora.bnb",]:
-        exec(f"import {x}", globals(), locals())
-        layers = dir(eval(x))
-        for fx in layers:
-            try: layer = eval(f"{x}.{fx}")
-            except: continue
-            if not hasattr(layer, "forward"): continue
-            if hasattr(eval(f"{x}.{fx}.forward"), "__wrapped__"): continue
-            exec(f"{x}.{fx}.forward = torch._disable_dynamo({x}.{fx}.forward)", globals(), locals())
+    import bitsandbytes
+    if Version(bitsandbytes.__version__) >= Version("0.46.0"):
+        if os.environ.get("UNSLOTH_ENABLE_LOGGING", "0") == "1":
+            print("Unsloth: Bitsandbytes >= 0.46.0 supports torch.compile - enabling.")
+    else:
+        # Disable dynamo on Linear4bit, Linear8bit and other future modules
+        if os.environ.get("UNSLOTH_ENABLE_LOGGING", "0") == "1":
+            print("Unsloth: Bitsandbytes < 0.46.0 does not support torch.compile - disabling.")
+        for x in ["bitsandbytes.nn.modules", "peft.tuners.lora.bnb",]:
+            exec(f"import {x}", globals(), locals())
+            layers = dir(eval(x))
+            for fx in layers:
+                try: layer = eval(f"{x}.{fx}")
+                except: continue
+                if not hasattr(layer, "forward"): continue
+                if hasattr(eval(f"{x}.{fx}.forward"), "__wrapped__"): continue
+                exec(f"{x}.{fx}.forward = torch._disable_dynamo({x}.{fx}.forward)", globals(), locals())
+            pass
         pass
     pass
 
@@ -90,9 +100,10 @@ def patch_torch_compile(debug = False, O3 = False, ignore_errors = True):
             graph_breaks = True,
             recompiles = True,
             recompiles_verbose = True,
-            compiled_autograd_verbose = True,
-            aot_joint_graph = True,
-            aot_graphs = True,
+            compiled_autograd_verbose = False, # Produces too much code
+            aot_joint_graph = False, # Produces too much code
+            aot_graphs = False,  # Produces too much code
+            perf_hints = True, # Performance improvement hints
         )
         torch._dynamo.config.verbose = True
     else:
@@ -153,6 +164,7 @@ def patch_torch_compile(debug = False, O3 = False, ignore_errors = True):
         f"config.cuda.compile_opt_level = {'-O2' if O3 else '-O1'}",
         # Capture torch.arange(...), torch.zeros(...)
         "config.capture_dynamic_output_shape_ops = True",
+        "config.emulate_precision_casts = False", # Force X.to(f32).to(f16) instead of X.to(f16)
     ]
     # Torch dynamo arguments
     torch_dynamo_arguments = [
@@ -165,6 +177,9 @@ def patch_torch_compile(debug = False, O3 = False, ignore_errors = True):
         # FAILS for Gemma!
         "config.compiled_autograd = False", # New Torch 2.4 feature which can compile backwards passes
         # https://pytorch.org/tutorials/intermediate/compiled_autograd_tutorial.html
+        "config.recompile_limit = 8", # Reduce recompiles to 8 - then will do eager
+        "config.allow_unspec_int_on_nn_module = True", # Integers in modules will auto wrap torch.tensor(self.vocab_size)
+        "config.optimize_ddp = True", # Optimizes DDP
     ]
     if not debug and ignore_errors:
         # Have to explicitly set it!
@@ -182,8 +197,45 @@ def patch_torch_compile(debug = False, O3 = False, ignore_errors = True):
     pass
 pass
 
+def get_model(model):
+    found_layers = False
+    x = model
+    while True:
+        if hasattr(x, "layers"):
+            found_layers = True
+            break
+        elif hasattr(x, "model"):
+            x = x.model
+        elif hasattr(x, "base_model"):
+            x = x.base_model
+        elif hasattr(x, "language_model"):
+            x = x.language_model
+        else:
+            break
+    pass
+    return x, found_layers
+pass
 
-def patch_model_and_tokenizer(model, tokenizer, downcast_rope = True, fix_embeddings = True):
+
+def verify_and_set_device(module,):
+    """
+    Verify that all parameters of a module are on the same device.
+    """
+    set_of_devices = set(x.device for x in module.parameters())
+    if len(set_of_devices) > 1:
+        raise ValueError(f"Unsloth: All parameters of {module} should be on the same device")
+    device = set_of_devices.pop()
+    module._per_layer_device_index = device.index
+pass
+
+def patch_model_and_tokenizer(
+    model,
+    tokenizer,
+    downcast_rope = True,
+    fix_embeddings = True,
+    do_forced_float32 = False,
+    correct_dtype = None,
+):
     # All Unsloth Zoo code licensed under LGPLv3
     assert(type(downcast_rope) is bool)
     import gc
@@ -217,15 +269,72 @@ def patch_model_and_tokenizer(model, tokenizer, downcast_rope = True, fix_embedd
     pass
 
     # Get most likely the correct data-type of the model
-    try:
-        correct_dtype = _get_dtype(model.config.torch_dtype)
-    except:
-        correct_dtype = model.get_input_embeddings().weight.dtype
+    if correct_dtype is None:
+        try:
+            correct_dtype = _get_dtype(model.config.torch_dtype)
+        except:
+            correct_dtype = model.get_input_embeddings().weight.dtype
+    pass
+    # If we force float32, we first use bfloat16, then downcast to float16
+    if do_forced_float32:
+      correct_dtype = torch.float16
+      for name, module in model.named_modules():
+          if "down_proj" in name or "up_proj" in name or "gate_proj" in name or "fc1" in name or "fc2" in name:
+              module.to(torch.float16)
+          if "q_proj" in name or "k_proj" in name or "v_proj" in name or "o_proj" in name or "out_proj" in name:
+              module.to(torch.float16)
+          if "lm_head" in name or "embed_tokens" in name:
+              module.to(torch.float16)
+          if "embed_tokens" in name or "patch_embedding" in name:
+              module.to(torch.float16)
+          if "norm" in name:
+              module.to(torch.float16)
+          torch.cuda.empty_cache()
+
+      # Convert any remaining bfloat16 parameters
+      for name, param in model.named_parameters():
+          if param.dtype == torch.bfloat16:
+              param.data = param.data.to(torch.float16)
+
+      # Also convert buffers (like position embeddings)
+      for name, buffer in model.named_buffers():
+          if buffer.dtype == torch.bfloat16:
+              buffer.data = buffer.data.to(torch.float16)
+      pass
+    pass
+
+    # Correct torch_dtype
+    def __fix_dtype(config):
+        if not hasattr(config, "to_dict"): return
+        dicts = config.to_dict()
+        for key, value in dicts.items():
+            if key == "torch_dtype":
+                setattr(config, "torch_dtype", correct_dtype)
+            else:
+                __fix_dtype(getattr(config, key))
+    m = model
+    while hasattr(m, "model"):
+        if hasattr(m, "dtype"):
+            try: setattr(m, "dtype", correct_dtype)
+            except: pass
+        if hasattr(m, "config"): __fix_dtype(m.config)
+        m = m.model
+    pass
+    if hasattr(m, "config"): __fix_dtype(m.config)
+    if hasattr(m, "dtype"):
+        try: setattr(m, "dtype", correct_dtype)
+        except: pass
+    pass
 
     # Check all params and patch!
     for name, module in model.named_modules():
         if isinstance(module, (Bnb_Linear4bit, Peft_Linear4bit)):
             weight = module.weight
+            # Check if quant_state exists for vision models like unsloth/Llama-3.2-11B-Vision-Instruct-bnb-4bit, unsloth/granite-vision-3.2-2b
+            if not hasattr(weight, 'quant_state'):
+                print(f"Skipping {name}: no quant_state found")
+                continue
+
             quant_state = weight.quant_state
 
             if type(quant_state) is list:
@@ -250,7 +359,7 @@ def patch_model_and_tokenizer(model, tokenizer, downcast_rope = True, fix_embedd
 
             elif hasattr(module, "short_cos_cached") and \
                 (module.short_cos_cached.dtype != correct_dtype):
-                
+
                 module.short_cos_cached = module.short_cos_cached.to(correct_dtype)
                 module.short_sin_cached = module.short_sin_cached.to(correct_dtype)
             pass
@@ -310,7 +419,7 @@ def patch_model_and_tokenizer(model, tokenizer, downcast_rope = True, fix_embedd
         lm_head.weight = old_output_embedding if not is_tied else old_input_embedding
         lm_head.in_features  = lm_head.weight.shape[1]
         lm_head.out_features = lm_head.weight.shape[0]
-        
+
         lm_head.weight.requires_grad_(requires_grad)
         model.set_output_embeddings(lm_head)
         if hasattr(model, "lm_head"): model.lm_head = lm_head
@@ -320,6 +429,13 @@ def patch_model_and_tokenizer(model, tokenizer, downcast_rope = True, fix_embedd
     # Otherwise error will occur on saving models ie use save_model
     if is_tied: model.tie_weights()
 
+    # For pipeline parallel models, we need to set the device for each layer for easier access later
+    x, found_layers = get_model(model)
+    if found_layers:
+        for layer in x.layers:
+            verify_and_set_device(layer)
+    pass
+
     # Clear deleted GPU items
     for _ in range(3):
         gc.collect()
@@ -344,7 +460,8 @@ def patch_compiled_autograd():
     source = source.split("\n")
     source = "\n".join(x[spaces:] for x in source)
     old = "return compiled_fn(inputs, sizes, scalars, hooks)"
-    n = len(re.search(r"\n([ ]{1,})return compiled_fn", source).group(1))
+    match = re.search(r"\n([ ]{1,})return compiled_fn", source)
+    n = len(match.group(1)) if match else 0
     source = source.replace(old, f"with disable():\n{' '*(n + 4)}{old}")
     source = source.replace("def end_capture", "def unsloth_end_capture", 1)
 
@@ -384,6 +501,95 @@ def patch_compiled_autograd():
 pass
 
 
+# utility function to help BC with old module hierarchy for transformers >=4.52.0
+def check_conversion_mappings(model, current_key_name_str, skip_modules):
+    # model_root_cls is None if there are no conversion_mappings or no _root_cls
+    model_root_cls = getattr(model, "_root_cls", model if hasattr(model, "_checkpoint_conversion_mapping") else None)
+    if model_root_cls is None:
+        return False
+    if hasattr(model_root_cls, "_checkpoint_conversion_mapping") and len(model_root_cls._checkpoint_conversion_mapping) > 0:
+        # if this is true, then it means that we must be on transformers >=4.52.0 because conversion_mappings was added in 4.52.0
+        # we cant know if the skip module naming convention is new or old
+        # but if we are supposed to skip this current_key_name_str, and it didn't pass
+        # (current_key_name_str in quantization_config.llm_int8_skip_modules)
+        # then new transformers + new module hierarchy means it should not be skipped, ie no BC check needed
+        # and new transformers + old module hierarchy means we still need to check to skip
+        # old transformers + old module hierarchy means no BC needed
+        # old transformers + new module hierarchy is problematic since we don't have the conversion_mappings to reverse
+        # follow the logic from save_pretrained in transformers.modeling_utils
+        reverse_conversion_mappings = {v: k for k, v in model_root_cls._checkpoint_conversion_mapping.items()}
+        new_current_key_names_str = current_key_name_str
+        for pattern, replacement in reverse_conversion_mappings.items():
+            try:
+                replacement = replacement.lstrip("^")  # strip off un-needed chars and patterns
+                replacement = re.sub(r"\(.*?\)", "", replacement)
+                key, n_replace = re.subn(pattern, replacement, current_key_name_str)
+                # Early exit of the loop
+                if n_replace > 0:
+                    new_current_key_names_str = key
+                    break
+            except Exception as e:
+                # skip this pattern but log
+                do_logging = os.environ.get('UNSLOTH_ENABLE_LOGGING', '0') == '1'
+                if do_logging:
+                    print(f"Unsloth: Replace bnb issue: {str(e)}")
+                break
+        return any([(skip_key + "." in new_current_key_names_str) or (skip_key == new_current_key_names_str) for skip_key in skip_modules])
+    return False
+
+
+def _mark_parent(child, parent_type):
+    """Attach the parent’s class so the child can inspect it later."""
+    child._root_cls = parent_type
+
+
+def _unmark_parent(child):
+    """Remove the temporary attribute if it is present."""
+    if hasattr(child, "_root_cls"):
+        delattr(child, "_root_cls")
+
+
+def parsed_statement(code: str) -> ast.stmt:
+    """Return the statement parsed from a one-liner."""
+    return ast.parse(code).body[0]
+
+
+class WrapRecursiveCall(ast.NodeTransformer):
+    function_name = "_replace_with_bnb_linear"
+    mark_statement = parsed_statement(
+        '_mark_parent(module, model._root_cls '
+        'if hasattr(model, "_root_cls") else type(model))'
+    )
+    unmark_statement = parsed_statement('_unmark_parent(module)')
+
+    def visit_Assign(self, node: ast.Assign):
+        """
+        Replace
+            _, has_been_replaced = _replace_with_bnb_linear(...)
+        with
+            try:
+                _mark_parent(module,
+                             model._root_cls
+                             if hasattr(model, "_root_cls")
+                             else type(model))
+                _, has_been_replaced = _replace_with_bnb_linear(...)
+            finally:
+                _unmark_parent(module)
+        """
+        if (
+            isinstance(node.value, ast.Call)
+            and getattr(node.value.func, "id", None) == self.function_name
+        ):
+            wrapped = ast.Try(
+                body      =[self.mark_statement, node],
+                handlers  =[],
+                orelse    =[],
+                finalbody =[self.unmark_statement],
+            )
+            return ast.copy_location(wrapped, node)
+        return node
+
+
 # Patch for dynamic 4bit quantization
 import inspect
 import transformers.integrations.bitsandbytes
@@ -399,18 +605,66 @@ if hasattr(transformers.integrations.bitsandbytes, "_replace_with_bnb_linear") a
     exec(f"from transformers.integrations.bitsandbytes import ({x})", globals())
     if "current_key_name_str" not in source:
         raise RuntimeError("Unsloth: Patch for dynamic quantization failed since current_key_name_str does not exist.")
-    
-    source = source.replace(
-        "name in quantization_config.llm_int8_skip_modules\n",
-        "((name in quantization_config.llm_int8_skip_modules) or (current_key_name_str in quantization_config.llm_int8_skip_modules))\n",
-        1,
-    )
 
-    # Need more than 1 replacement since recursion is done
+    # First patch recursive calls to mark the parent class
+    # we need it to access the parent class to check for conversion_mappings
+    try:
+        mark_parent_error = False
+        new_source = source.replace(
+            "name in quantization_config.llm_int8_skip_modules\n",
+            "((name in quantization_config.llm_int8_skip_modules) or (current_key_name_str in quantization_config.llm_int8_skip_modules) or (check_conversion_mappings(model, current_key_name_str, quantization_config.llm_int8_skip_modules)))\n",
+            1,
+        )
+
+        source_tree = ast.parse(new_source)
+        source_tree = WrapRecursiveCall().visit(source_tree)
+        ast.fix_missing_locations(source_tree)
+        new_source = ast.unparse(source_tree)
+
+        # will raise error if patch fails
+        compile(new_source, '<temp_patched>', 'exec')
+        if '_mark_parent' not in new_source and '_unmark_parent' not in new_source:
+            do_logging = os.environ.get('UNSLOTH_ENABLE_LOGGING', '0') == '1'
+            if do_logging:
+                print(f"Unsloth: Could not wrap replace_with_bnb_linear but may not be an issue")
+            mark_parent_error = True
+        else:
+            source = new_source
+
+    except Exception as e:
+        do_logging = os.environ.get('UNSLOTH_ENABLE_LOGGING', '0') == '1'
+        if do_logging:
+            print(f"Unsloth: Could not wrap replace_with_bnb_linear but may not be an issue. {str(e)}")
+        mark_parent_error = True
+
+    if mark_parent_error:
+        # we sitll have the original source without the mark_parent and unmark_parent patches
+        source = source.replace(
+            "name in quantization_config.llm_int8_skip_modules\n",
+            "((name in quantization_config.llm_int8_skip_modules) or (current_key_name_str in quantization_config.llm_int8_skip_modules))\n",
+            1,
+        )
+
     source = source.replace(
         "_replace_with_bnb_linear",
         "_unsloth_replace_with_bnb_linear",
     )
+
+    score_code = """if name == 'score':
+    modules_to_not_convert.append("score")"""
+
+    pattern = r"(^\s*)(current_key_name\.append\(name\))"
+
+    def add_score_code(match):
+        indentation = match.group(1)  # Captured indentation
+        line_content = match.group(2) # The line 'current_key_name.append(name)'
+
+        indented_breakpoint_code = "\n".join([f"{indentation}{line}" for line in score_code.splitlines()])
+
+        return f"{indentation}{line_content}\n{indented_breakpoint_code}"
+
+    source = re.sub(pattern, add_score_code, source, flags=re.MULTILINE)
+
     exec(source, globals())
     transformers.integrations.bitsandbytes._replace_with_bnb_linear = _unsloth_replace_with_bnb_linear
 pass
