@@ -60,7 +60,7 @@ import base64
 from io import BytesIO
 import math
 import requests
-from typing import Union, Tuple
+from typing import Union, Tuple, List, Dict
 IMAGE_FACTOR = 28
 MIN_PIXELS = 4 * 28 * 28
 MAX_PIXELS = 16384 * 28 * 28
@@ -93,7 +93,7 @@ pass
 
 def smart_resize(
     height: int, width: int, factor: int = IMAGE_FACTOR, min_pixels: int = MIN_PIXELS, max_pixels: int = MAX_PIXELS
-) -> tuple[int, int]:
+) -> Tuple[int, int]:
     """
     Rescales the image so that the following conditions are met:
 
@@ -122,7 +122,7 @@ pass
 
 
 def fetch_image(
-    ele: dict[Union[Tuple[str, str], Image.Image]],
+    ele: Dict,
     size_factor: int = IMAGE_FACTOR,
 ) -> Image.Image:
     if "image" in ele:
@@ -173,7 +173,7 @@ def fetch_image(
 pass
 
 
-def extract_vision_info(conversations: Union[list[dict], list[list[dict]]]) -> list[dict]:
+def extract_vision_info(conversations: Union[List[Dict], List[List[Dict]]]) -> List[Dict]:
     vision_infos = []
     if isinstance(conversations[0], dict):
         conversations = [conversations]
@@ -181,20 +181,15 @@ def extract_vision_info(conversations: Union[list[dict], list[list[dict]]]) -> l
         for message in conversation:
             if isinstance(message["content"], list):
                 for ele in message["content"]:
-                    if (
-                        "image" in ele
-                        or "image_url" in ele
-                        or "video" in ele
-                        or ele["type"] in ("image", "image_url", "video")
-                    ):
+                    if ele["type"] in ("image", "image_url", "video"):
                         vision_infos.append(ele)
     return vision_infos
 pass
 
 
 def process_vision_info(
-    conversations: Union[list[dict], list[list[dict]]],
-) -> tuple[Union[list[Image.Image], None], Union[list[Union[torch.Tensor, list[Image.Image]]], None]]:
+    conversations: Union[List[Dict], List[List[Dict]]],
+) -> Tuple[Union[List[Image.Image], None], Union[List[Union[torch.Tensor, List[Image.Image]]], None]]:
     vision_infos = extract_vision_info(conversations)
     ## Read images or videos
     image_inputs = []
@@ -261,13 +256,14 @@ class UnslothVisionDataCollator:
     __slots__ = \
         "padding_token_ids", "dtype", "ignore_index", \
         "processor", "formatting_func", "image_size", \
-        "max_seq_length", "truncation", "train_on_responses_only",
+        "max_seq_length", "truncation", "train_on_responses_only", \
+        "num_proc", "assistant_single_content",
 
     def __init__(
         self,
         model,
         processor,
-        max_seq_length = None,
+        max_seq_length  = None,
         formatting_func = None,
         resize = "min", # Can be (10, 10) or "min" to resize to fit
                         # the model's default image_size or "max"
@@ -277,6 +273,7 @@ class UnslothVisionDataCollator:
         instruction_part = None,
         response_part    = None,
         force_match      = True, # Match newlines as well!
+        num_proc         = None,
     ):
         if not hasattr(processor, "image_processor"):
             raise TypeError("Unsloth: UnslothVisionDataCollator is only for image models!")
@@ -329,9 +326,40 @@ class UnslothVisionDataCollator:
                 force_match      = force_match,
                 tokenizer        = processor,
                 return_function  = True,
+                num_proc         = num_proc,
             )
         else:
             self.train_on_responses_only = None
+
+        # Check what type for assistant VLM tokenizer allows!
+        # Good for Mistral V3 and Pixtral I think
+        try:
+            processor.apply_chat_template([
+                {"role": "user", "content": [
+                    {"type": "image"},
+                    {"type": "text", "text": "Hello!"}]},
+                {"role": "assistant", "content": [
+                    {"type": "text", "text": "How can I help you?"}]}
+            ])
+            self.assistant_single_content = False
+        except TypeError:
+            try:
+                processor.apply_chat_template([
+                    {"role": "user", "content": [
+                        {"type": "image"},
+                        {"type": "text", "text": "Hello!"}]},
+                    {"role": "assistant", "content": "How can I help you?"}
+                ])
+                self.assistant_single_content = True
+                print(
+                    f"Unsloth: {processor.__class__.__name__} only accepts 1 "\
+                    "text field for assistant roles!\n"\
+                    "We will auto fix the data collator to support it!"
+                )
+            except Exception as e:
+                raise RuntimeError(e)
+        except Exception as e:
+            raise RuntimeError(e)
         return
     pass
 
@@ -363,7 +391,7 @@ class UnslothVisionDataCollator:
                     )
                 content = message["content"]
                 if type(content) is str:
-                    message["content"] = [{"type" : "text", "text" : content}]
+                    message["content"] = content = [{"type" : "text", "text" : content}]
                 elif type(content) is list or type(content) is tuple:
                     part = content[0]
                     assert("type" in part)
@@ -374,6 +402,15 @@ class UnslothVisionDataCollator:
                         "[{'role':'user', 'content':[{'type':'text', 'text':'Hello!'}]}]"
                     )
                 pass
+
+                # Also fix the messages if assistant must only be 1 string!
+                # Only affects Mistral V3 I think!
+                if self.assistant_single_content:
+                    for message in messages:
+                        if message["role"] == "assistant":
+                            if type(content := message["content"]) is list:
+                                message["content"] = content[0]["text"]
+                pass
             pass
             message = self.processor.apply_chat_template(
                 messages,
@@ -414,7 +451,8 @@ class UnslothVisionDataCollator:
             return_tensors = "pt",
             add_special_tokens = False, # Stop double BOS
         )
-        batch.pop("token_type_ids", None)
+        # Cannot remove due to bidirectional attention from Gemma 3!
+        # batch.pop("token_type_ids", None)
 
         # Pixtral accepts multiple images, so we have to cast it individually
         pixel_values = batch["pixel_values"]
@@ -435,7 +473,6 @@ class UnslothVisionDataCollator:
         labels = batch["input_ids"].clone()
         labels[torch.isin(labels, self.padding_token_ids)] = self.ignore_index
         batch["labels"] = labels
-
         if self.train_on_responses_only:
             batch["labels"] = self.train_on_responses_only(batch)["labels"]
         return batch
